package f00f.net.irc.martyr;

import java.util.Observable;
import java.util.Observer;

import f00f.net.irc.martyr.errors.NickInUseError;

import f00f.net.irc.martyr.commands.NickCommand;
import f00f.net.irc.martyr.commands.UserCommand;

import f00f.net.irc.martyr.clientstate.ClientState;
import f00f.net.irc.martyr.util.FullNick;

/**
 * <p>Receiving an ERR_NICKNAMEINUSE while unregistered will cause the
 * auto-register to try again, with an _ appended to the nick.  If
 * this fails five times, AutoRegister will ask the IRCConnection to
 * disconnect().  Note that if it fails to connect it remains enabled, so that
 * if IRCConnection.connect() is called, it will re-try the same 5 NICKs.</p>
 *
 * <p>AutoRegister will add itself as a state observer and as a command
 * observer.  It needs to receive the error.NickInUseError command so that
 * it can re-try the registration, and it needs to detect when we
 * transition into the UNREGISTERED state.</p>
 *
 * <p>AutoRegister should be created before the IRCConnection.connect()
 * is called.  AutoRegister can be disabled by calling the 'disable()'
 * method at any time.  This simply removed AutoRegister as an
 * observer for the state and commands.</p>
 *
 */
public class AutoRegister extends GenericAutoService
{

private String originalNick;
private String nick = null;
private String user;
private String name;
private int attempts = 0;

public static final int MAX_ATTEMPTS = 5;

public AutoRegister( IRCConnection connection, String nick,
	String user, String name )
{
	super( connection );

	this.originalNick = nick;
	this.user = user;
	this.name = name;

	enable();
}

protected void updateState( State state )
{
	if( state == State.UNCONNECTED )
	{
		try
		{
			// This should only happen if we were connected and then
			// disconnected.  Try connecting.
			ClientState cstate = getConnection().getClientState();
			getConnection().connect( cstate.getServer(), cstate.getPort() );
		}
		catch( Exception e )
		{
			// Ok, we give up.
		}
	}
	else if( state == State.UNREGISTERED )
	{
		// We need to do some registerin'!
		attempts = 0;
		nick = originalNick;
		doRegister();
	}
	else if( state == State.REGISTERED )
	{
		// we need to update the client state.
		ClientState clientState = getConnection().getClientState();
		clientState.setNick( new FullNick( nick ) );
		clientState.setName( name );
		clientState.setUser( user );
	}
}

protected void updateCommand( Command command )
{
	if( command instanceof NickInUseError &&
		getConnection().getState() == State.UNREGISTERED )
	{
		++attempts;
		if( attempts >= MAX_ATTEMPTS )
		{
			Debug.println( this, "Failed to register.", Debug.NORMAL );
			getConnection().disconnect();
			return;
		}
		
		// Append _ to the nick, re-register.
		nick = nick + '_';
		doRegister();
	}
}

private void doRegister()
{
	if( getConnection().getState() != State.UNREGISTERED )
	{
		Debug.println( this,
			"Tried to register but we are not UNREGISTERED", Debug.FAULT );
		return;
	}

	Debug.println( this, "Trying to register as " + nick, Debug.NORMAL );

	getConnection().sendCommand( new NickCommand( nick ) );
	getConnection().sendCommand( new UserCommand( user, name, getConnection() ) );
}

public String toString()
{
	return "AutoRegister [" + attempts + "]";
}

// END AutoRegister
}
 


 
