/*
 * IRCConnection.java
 *
 *   Copyright (C) 2000, Ben Damm
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   See: http://www.fsf.org/copyleft/lesser.txt
 */

package f00f.net.irc.martyr;

import f00f.net.irc.martyr.commands.UnknownCommand;

import f00f.net.irc.martyr.clientstate.ClientState;

import java.io.IOException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Observer;
import java.util.Observable;
import java.util.StringTokenizer;

/**
 * <p>IRCConnection is the core class for Martyr.  IRCConnection manages
 * the socket, giving commands to the server and passing results to
 * the parse engine.</p>
 *
 * <p>Registration is left up to the consumer.  Registration should occur
 * any time the state changes to 'UNREGISTERED'.  The consumer will
 * know this because it has registered some class as a state observer.
 * The class 'AutoRegister' can do this automatically.</p>
 *
 * <p>Some commands, such as Ping require an automatic response.
 * Commands that fall into this category (currently, only PING) can be
 * handled by the 'AutoResponder' class.  </p>
 *
 * <p>The ClientStateMonitor class tells commands to change the client
 * state when they are received.  ClientStateMonitor is automatically
 * added to the command queue before any other command, so that you
 * can be guaranteed that the ClientState is updated before any
 * observer sees a command.</p>
 *
 * <p>You can use the AutoJoin class to join a channel and stay there.
 * AutoJoin will try to re-join if kicked or if the connection is lost
 * and the server re-connects.  AutoJoin can be used any time a join
 * is desired.  If the server is not connected, it will wait until the
 * server connects.  If the server is connected, it will try to join
 * right away.</p>
 *
 *
 * <p>You will probably want to use the AutoRegister and AutoResponder classes.
 * Example:</p>
 *
 * <p>Note that in the example, the first line is optional. IRCConnection can be
 * called with its default constructor.  See note below about why this is done.
 * IRCConnection will instantiate its own ClientState object if you do not
 * provide one.</p>
 * 
 * <pre>
 * 
 * ClientState clientState = new MyAppClientState();
 * IRCConnection connection = new IRCConnection( clientState );
 *
 * // AutoRegister and AutoResponder both add themselves to the
 * // appropriate observerables.  Both will remove themselves with the
 * // disable() method.
 * 
 * AutoRegister autoReg
 *   = new AutoRegister( "repp", "bdamm", "Ben Damm", connection );
 * AutoResponder autoRes = new AutoResponder( connection );
 *
 * // Very important that the guys above get added before the connect.
 * // If done otherwise, AutoRegister should throw an
 * // IllegalStateException.  Otherwise, AutoRegister can't catch the
 * // state switch to UNREGISTERED from UNCONNECTED.
 *
 * connection.connect( server, port );
 * </pre>
 *
 * <p>So, how does an application know when a channel has been joined, a user has
 * joined a channel we are already on, etc?  How does the application get
 * fine-grained access to client state change info?  This is a tricky question,
 * and the current solution is to sublcass the clientstate.ClientState
 * and clientstate.Channel classes with your own, overriding the
 * setXxxxXxxx methods.  Each method would call super.setXxxXxxx and
 * then proceed to change the application as required.</p>
 *
 * <p>IRCConnection starts in the UNCONNECTED state and makes no attempt
 * to connect until the connect(...) method is called.</p>
 * @see f00f.net.irc.martyr.clientstate.ClientState
 */
public class IRCConnection {

/** State of the session. */
private State state;

/**
 * Client state (our name, nick, groups, etc). Stored here mainly
 * because there isn't anywhere else to stick it.
 */
private ClientState clientState;

/**
 * Maintains a list of classes observing the state and notifies them
 * when it changes.
 */
private StateObserver stateObservers;

/**
 * Maintains a list of classes observing commands when they come in.
 */
private CommandObserver commandObservers;

/**
 * The actual socket used for communication.
 */
private Socket socket;

/**
 * Object to use as a monitor for the socketWriter.
 */
private Object sendingMonitor = new Object();

/**
 * The writer to use for output.
 */
private BufferedWriter socketWriter;

/**
 * The reader to use for input.  Managed by the InputManager.
 */
private BufferedReader socketReader;

/**
 * Command register, contains a list of commands that can be received
 * by the server and have matching Command objects.
 */
private CommandRegister commandRegister;

/**
 * Maintains a handle on the input handler.
 */
private Thread inputHandler;

private boolean disconnecting = false;
private boolean connected = false;

public IRCConnection()
{
	this( new ClientState() );
}

public IRCConnection( ClientState clientState )
{
	stateObservers = new StateObserver();
	commandObservers = new CommandObserver();
	this.clientState = clientState;
	
	commandRegister = new CommandRegister();

	setState( State.UNCONNECTED );

	new ClientStateMonitor( this );
}

public void connect( String server, int port )
	throws UnknownHostException, IOException
{
	if( state != State.UNCONNECTED && ! disconnecting )
		return;

	connected = false;
	
	socket = new Socket( server, port );

	socketWriter =
		new BufferedWriter( new OutputStreamWriter(	socket.getOutputStream() ) );
	
	socketReader = 
		new BufferedReader( new InputStreamReader( socket.getInputStream() ) );

	// A simple thread that waits for input from the server and passes
	// it off to the IRCConnection class.
	inputHandler = new InputHandler( socketReader, this );
	inputHandler.start();

	clientState.setServer( server );
	clientState.setPort( port );

	// This is to avoid sending out an UNREGISTERED state switch while
	// some state listener asked us to connect.  The rest of the state
	// listeners would get UNREGISTERED and then UNCONNECTED instead
	// of UNCONNECTED and then UNREGISTERED.  If we are disconnecting,
	// disconnect() will handle the state switch.
	connected = true;
	if( ! disconnecting )
		setState( State.UNREGISTERED );
}

/**
 * Orders the socket to disconnect.
 */
public void disconnect()
{
	if( disconnecting || !connected )
		return;
	disconnecting = true;
	
	Debug.println( this, "Stopping input handler.", Debug.EXCESSIVE );
	// Deprecated?
	// inputHandler.stop();

	Debug.println( this, "Closing socket.", Debug.EXCESSIVE );
	try
	{
		socket.close();
	}
	catch( IOException ioe )
	{
		// And we are supposed to do what?
		Debug.println( this, ioe.toString(), Debug.CRITICAL );
	}

	connected = false;

	// It is important that the state be switched last.  One of the
	// state listeners may try to re-connect us.
	setState( State.UNCONNECTED );

	disconnecting = false;

	// And in case we were re-connected...
	if( connected )
		setState( State.UNREGISTERED );
}

public void setDaemon( boolean daemon )
{
	inputHandler.setDaemon( daemon );
}

public String toString()
{
	return "IRCConnection";
}

public void addStateObserver( Observer observer )
{
	Debug.println( this, "Added state observer " + observer, Debug.EXCESSIVE );
	stateObservers.addObserver( observer );
}

public void removeStateObserver( Observer observer )
{
	Debug.println( this, "Removed state observer " + observer, Debug.EXCESSIVE );
	stateObservers.deleteObserver( observer );
}

public void addCommandObserver( Observer observer )
{
	Debug.println( this, "Added command observer " + observer, Debug.EXCESSIVE );
	commandObservers.addObserver( observer );
}

public void removeCommandObserver( Observer observer )
{
	Debug.println( this, "Removed command observer " + observer, Debug.EXCESSIVE );
	commandObservers.deleteObserver( observer );
}


public State getState()
{
	return state;
}

public ClientState getClientState()
{
	return clientState;
}

public void sendCommand( Command command )
{
	sendCommand( command.render() );
}

public String getLocalhost()
{
	return "localhost";
}

public String getRemotehost()
{
	return clientState.getServer();
}

// ===== private methods =============================================

private void setState( State newState )
{
	if( state == newState )
		return;

	state = newState;

	Debug.println( this, "State switch: " + state, Debug.VERBOSE );

	stateObservers.setChanged();
	stateObservers.notifyObservers( newState );
}

/**
 * Sends the command down the socket, with the required 'CRLF' on the
 * end.
 */
private void sendCommand( String str )
{
	try
	{
		synchronized( sendingMonitor )
		{
			try
			{
				// Slow things down a bit so the server doesn't kill us
				Thread.sleep(300);
			}
			catch( InterruptedException ie )
			{
			}
	
			Debug.println( this, "SEND= " + str, Debug.NORMAL );
			socketWriter.write( str + "\r\n" );
			socketWriter.flush();
		}
	}
	catch( IOException ioe )
	{
		socketError( ioe );
	}
}

// ===== package methods =============================================

void socketError( IOException ioe )
{
	Debug.println( this, ioe.toString(), Debug.CRITICAL );
	disconnect();
}

synchronized void incomingCommand( String wholeString )
{
	Debug.println( this, "RCV = " + wholeString, Debug.NORMAL );

	// 1) Parse out the command
	String prefix = "";
	String identifier = "";
	String params = "";
 
	StringTokenizer tokens = new StringTokenizer( wholeString, " " );
 
	if( wholeString.charAt(0) == ':' )
	{
		prefix = tokens.nextToken();
		prefix = prefix.substring( 1, prefix.length() );
	}
 
	identifier = tokens.nextToken();
 
	if( tokens.hasMoreTokens() )
	{
		// The rest of the string
		params = tokens.nextToken("");
	}
	
	// 2) Fetch command from factory
	// Remember that commands are also factories.
	Command commandFactory = commandRegister.getCommand( identifier );
	Command command;

	if( commandFactory == null )
	{
		// The identifier doesn't map to a command.
		Debug.println( this, "Unknown command", Debug.CRITICAL );
		command = new UnknownCommand();
	}
	else
	{
		command = commandFactory.parse( prefix, identifier, params);
	
		if( command == null )
		{
			Debug.println( this,"Command factory returned NULL", Debug.CRITICAL );
			return;
		}
		Debug.println( this, "Using " + command, Debug.EXCESSIVE );
	}


	command.setSourceString( wholeString );

	// 3) Change the connection state if required
	// This allows us to change from UNREGISTERED to REGISTERED and
	// possibly back.
	State cmdState = command.getState();
	if( cmdState != State.UNKNOWN && cmdState != getState() )
		setState( cmdState );

	// 4) Notify command observers
	commandObservers.setChanged();
	commandObservers.notifyObservers( command );
}

// ----- END IRCConnection -------------------------------------------
}

/**
 * A simple class to help manage input from the stream.
 */
class InputHandler extends Thread
{

private BufferedReader reader;
private IRCConnection connection;

public InputHandler( BufferedReader reader, IRCConnection connection )
{
	this.reader = reader;
	this.connection = connection;
}

public void run()
{
	try{
		
		String str;
		while( true )
		{
			str = reader.readLine();
			if( str == null )
			{
				connection.socketError( new IOException( "Socket disconnected" ) );
				return;
			}
			connection.incomingCommand( str );
		}
	}
	catch( IOException ioe )
	{
		connection.socketError( ioe );
	}
}

// ----- END InputHandler --------------------------------------------
}

/**
 * Should the state and state observer be one?
 */
class StateObserver extends Observable
{
	public void setChanged()
	{
		super.setChanged();
	}
}

class CommandObserver extends StateObserver
{
}


