package f00f.net.irc.martyr.clientstate;

import java.util.Hashtable;
import java.util.Enumeration;

import f00f.net.irc.martyr.clientstate.Channel;
import f00f.net.irc.martyr.util.FullNick;

/**
 * <p>Maintains a list of client-related facts such as what channels we
 * are in, who else is in the channels, what our nick is, etc.</p>
 *
 * <p>ClientState is a critical part of martyr.  To get access to events
 * that change the client state, the framework user can subclass
 * ClientState and then pass the subclass to the IRCConnection's
 * constructor.  Then, when a command detects a change in client
 * state, it will call the corresponding method in the custom
 * ClientState.</p>
 *
 * <p>If a user of the framework wishes to grab client state information
 * about a channel (when a user joins, when a user leaves, topic
 * change, etc), the user can do so in a similar manner.  Simply
 * override the 'addChannel(String)' method to instantiate their own
 * Channel subclass, and add it to the protected 'channels' hastable,
 * similar to this:</p>
 *
 * <pre>
 * public void addChannel( String channame )
 * {
 *     channels.put( channame, new Channel( channel ) );
 * }
 * </pre>
 *
 */
public class ClientState
{

private FullNick nick = null;
private String user = "";
private String name = "";
private String server = "";
private int port = -1;

// Hashtable is threadsafe so we don't have to be.
protected Hashtable channels = new Hashtable();

public void setNick( FullNick nick )
{
	this.nick = nick;
}

public FullNick getNick()
{
	return nick;
}

public void setUser( String user )
{
	this.user = user;
}

public void setName( String name )
{
	this.name = name;
}

public void setServer( String server )
{
	this.server = server;
}

public String getServer()
{
	return server;
}

public void setPort( int port )
{
	this.port = port;
}

public int getPort()
{
	return port;
}

public void addChannel( String channame )
{
	channels.put( channame, new Channel( channame ) );
}

public Channel getChannel( String chanName )
{
	return (Channel)channels.get( chanName );
}

/**
 * Removes a channel from the state, does nothing if the channel name
 * is invalid.
 * Should we throw an exception here?
 */
public void removeChannel( String channel )
{
	channels.remove( channel );
}

public Enumeration getChannelNames()
{
	return channels.keys();
}

}


